#!/usr/bin/env python
# -*- coding: utf-8 -*-

"Collection of objects related by the central dogma of biology."

__version__ = '0.0.1'
name = 'dogma'


from dogma.nbci import (
    NBCI_GENETIC_CODE_NAMES,
    NCBI_NUCLEOTIDES,
    NCBI_CODONS,
    NCBI_AMINO_ACIDS,
)

from dogma.utils import (
    DEFAULT_DECIMAL_PRECISION,
    STANDARD_NUCLEOTIDES,
    DEGENERATE_NUCLEOTIDE_CODE,
    DEGENERATE_NUCLEOTIDES,
    DEGENERATE_NUCLEOTIDE_CODE_REVERSED,
    DEGENERATE_NUCLEOTIDE_CODE_COMPOSITION,
    NUCLEOTIDE_BASE_PAIRS,
    DEGENERATE_NUCLEOTIDE_PAIRS,
    DEFAULT_NUCLEOTIDE_LABEL,
    DEFAULT_OLIGONUCLEOTIDE_LABEL,
    STANDARD_CODONS,
    DEGENERATE_CODONS,
    DEFAULT_CODON_LABEL,
    STANDARD_AMINO_ACIDS,
    DEFAULT_AMINO_ACID_LABEL,
    STOP_LABEL,
    DEFAULT_RESIDUE_LABEL,
    rescale,
    get_frequency_dictionary,
    get_random_oligonucleotide
)

from dogma.genetic_codes import (
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    translate)

from dogma.nucleotides import (
    Nucleotide,
    combine_nucleotides,
    is_valid_nucleotide_string
)

from dogma.codons import (
    Codon,
    degenerate_codon_string_to_standard_members,
    nucleotide_string_to_codons,
    combine_codons,
    combine_codon_labels,
    merge_codons
)

from dogma.oligonucleotides import (
    Oligonucleotide,
    reverse_complement
)

from dogma.amino_acids import AminoAcid

from dogma.proteins import (
    Protein,
    calculate_protein_degeneracy
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Translation system information from NCBI.

:Reference:
    Nomenclature for Incompletely Specified Bases in Nucleic Acid Sequences
    http://www.sbcs.qmul.ac.uk/iubmb/misc/naseq.html
"""

from itertools import product


# dictionary of NCBI genetic code ID numbers and full names
NBCI_GENETIC_CODE_NAMES = {
    1: 'Standard',
    2: 'Vertebrate Mitochondrial',
    3: 'Yeast Mitochondrial',
    4: 'Mold Mitochondrial; Protozoan Mitochondrial; Coelenterate;'
       'Mitochondrial; Mycoplasma; Spiroplasma',
    5: 'Invertebrate Mitochondrial',
    6: 'Ciliate Nuclear; Dasycladacean Nuclear; Hexamita Nuclear',
    9: 'Echinoderm Mitochondrial; Flatworm Mitochondrial',
    10: 'Euplotid Nuclear',
    11: 'Bacterial, Archaeal and Plant Plastid',
    12: 'Alternative Yeast Nuclear',
    13: 'Ascidian Mitochondrial',
    14: 'Alternative Flatworm Mitochondrial',
    15: 'Blepharisma Macronuclear',
    16: 'Chlorophycean Mitochondrial',
    21: 'Trematode Mitochondrial',
    22: 'Scenedesmus obliquus Mitochondrial',
    23: 'Thraustochytrium Mitochondrial',
    24: 'Pterobranchia Mitochondrial',
    25: 'Candidate Division SR1 and Gracilibacteria',
    26: 'Pachysolen tannophilus Nuclear Code'
}

# string of one-letter codes for the four standard nucleotides
NCBI_NUCLEOTIDES = 'ACGT'

# list of all 64 standard codons ['AAA', 'AAC', ... 'TTG', 'TTT']
NCBI_CODONS = [''.join(_) for _ in product(NCBI_NUCLEOTIDES, repeat=3)]

# string of one-letter amino acid codes aligned with alphabetized codon list
NCBI_AMINO_ACIDS = {
    1: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSS*CWCLFLF',
    2: 'KNKNTTTT*S*SMIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    3: 'KNKNTTTTRSRSMIMIQHQHPPPPRRRRTTTTEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    4: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    5: 'KNKNTTTTSSSSMIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    6: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVVQYQYSSSS*CWCLFLF',
    9: 'NNKNTTTTSSSSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    10: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSCCWCLFLF',
    11: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSS*CWCLFLF',
    12: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLSLEDEDAAAAGGGGVVVV*Y*YSSSS*CWCLFLF',
    13: 'KNKNTTTTGSGSMIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    14: 'NNKNTTTTSSSSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVVYY*YSSSSWCWCLFLF',
    15: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*YQYSSSS*CWCLFLF',
    16: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*YLYSSSS*CWCLFLF',
    21: 'NNKNTTTTSSSSMIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    22: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*YLY*SSS*CWCLFLF',
    23: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSS*CWC*FLF',
    24: 'KNKNTTTTSSKSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSWCWCLFLF',
    25: 'KNKNTTTTRSRSIIMIQHQHPPPPRRRRLLLLEDEDAAAAGGGGVVVV*Y*YSSSSGCWCLFLF',
    26: 'FFLLSSSSYY**CC*WLLLAPPPPHHQQRRRRIIIMTTTTNNKKSSRRVVVVAAAADDEEGGGG'
}



import decimal
import math

import pandas as pd

from dogma import (
    DEFAULT_DECIMAL_PRECISION,
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    translate,
    Nucleotide,
    Codon,
)


decimal.getcontext().prec = DEFAULT_DECIMAL_PRECISION


class Oligonucleotide:
    """
    A sequence of Nucleotides objects.
    5' --> 3' list
    """

    def __init__(self, data, genetic_code=None, auto_run=True):
        """
        Multiple input formats are acceptable
            Oligonucleotide('AAANNK')
            Oligonucleotide(list('AAANNK'))
            Oligonucleotide(list(map(Nucleotide, list('AAANNK'))))
            Oligonucleotide([{'A':1}, {'A':0.5, 'C':0.5}]
            Oligonucleotide(list(map(Nucleotide, list('AAANNK'))))
        """

        if not isinstance(genetic_code, GeneticCode):
            genetic_code = DEFAULT_GENETIC_CODE
        self.genetic_code = genetic_code

        if isinstance(data, (str, list)):
            if isinstance(data[0], Codon):
                self.bases = []
                for c in data:
                    self.bases += c.bases
            else:
                self.bases = [_
                              if isinstance(_, Nucleotide)
                              else Nucleotide(_)
                              for _ in data]
        elif isinstance(data, Codon):
            self.bases = data.bases
            self.genetic_code = data.genetic_code
        else:
            self.bases = None

        self.length = self.get_length()
        self.label = self.get_label()
        self.compact_label = self.get_compact_label()

        self.codon_strings = self.get_codon_strings()
        self.codons = self.get_codons()

        self.get_base_profile()
        self.get_codon_profile()
        self.get_amino_acid_profile()

        if auto_run:
            self.assess_degeneracy()

    def get_bases(self):
        """
        Returns list of Nucleotides objects
        """
        return self.bases

    def get_codons(self):
        """
        Returns list of Codon objects derived from oligonucleotide bases
        """
        base_triplets = [self.bases[3 * i: 3 * i + 3]
                         for i in range(self.length // 3)]
        return [Codon(_, self.genetic_code) for _ in base_triplets]

    def get_length(self):
        """
        Returns integer length of oligonucleotide, number of bases
        """
        return len(self.bases)

    def get_label(self):
        """
        Returns a string of nucleotide labels.
        """
        return ''.join(_.label for _ in self.bases)

    def get_codon_strings(self):
        """
        Returns list of codon strings.
        """
        return [self.label[3 * i: 3 * i + 3] for i in range(self.length // 3)]

    def get_compact_label(self):
        """
        Returns shorthand label
        Attempts to reduce identical codon labels with numbers
        Ex: NNKNNKNNKAAANNK --> NNK<3>AAANNK
        """
        count = 1
        previous_string = ''
        output = []
        for c in self.get_codon_strings():
            if c == previous_string:
                count += 1
            else:
                if count > 1:
                    output.append(f'<{count}>')
                previous_string = c
                output.append(c)
                count = 1
        if count > 1:
            output.append(f'<{count}>')
        return ''.join(output)

    def translate(self, genetic_code=None):
        """
        Translates oligonucleotide into protein.

        Parameters:
        genetic_code: GeneticCode object
        """
        if not isinstance(genetic_code, GeneticCode):
            genetic_code = self.genetic_code
        return ''.join([genetic_code[_] for _ in self.get_codon_strings()])

    def samples(self, k=1, output='oligonucleotide', by_codon=True):
        return [self.sample(output=output, by_codon=by_codon)
                for _ in range(k)]

    def sample(self, output='oligonucleotide', by_codon=True):
        """
        Returns a single non-degenerate oligonucleotide string.

        If by_codon is True, sampling is based on the codons attribute,
        otherwise sampling is performed on each base individually.

        """
        if by_codon:
            oligonucleotide_string = ''.join(c.sample() for c in self.codons)
        else:
            oligonucleotide_string = ''.join([b.sample() for b in self.bases])

        if output == 'protein':
            return translate(oligonucleotide_string, self.genetic_code)
        elif output == 'both':
            return oligonucleotide_string, translate(oligonucleotide_string,
                                                     self.genetic_code)
        else:
            return oligonucleotide_string

    def assess_degeneracy(self):
        """
        Performs the sequential and potentially computationally and memory
        intensive calculations related to generating the protein abundance
        profile.
        """

        self.get_amino_acid_degeneracy_profile()
        self.get_protein_degeneracy_table()
        self.get_degeneracy_table()
        self.get_size()

    def get_base_profile(self):
        """
        Generates and returns base profile

        Usage
        -----
        >> nnk = Codon('NNK')
        >> oligo = Oligonucleotide(nnk)
        >> oligo.get_base_profile()
            {0: {'A':1, 'C':1 , 'G':1. 'T':1},
             1: {'A':1, 'C':1 , 'G':1. 'T':1},
             2: {               'G':1. 'T':1}}
        """
        self.base_profile = [b.composition for b in self.bases]
        return self.base_profile

    def get_codon_profile(self):
        """
        Generates and returns codon profile

        Usage
        -----
        >> nnk = Codon('NNK')
        >> oligo = Oligonucleotide(nnk)
        >> oligo.get_codon_profile()
            {0: {'AAA':1, 'AAC':1, AAG':1, 'AAT':1, 'ACA':1, ...},
             1: {'AAA':1, 'AAC':1, AAG':1, 'AAT':1, 'ACA':1, ...},
             2: {                 'AAG':1, 'AAT':1,          ...}}

        """
        self.codon_profile = [c.composition for c in self.codons]
        return self.codon_profile

    def get_amino_acid_profile(self):
        """
        Generates and returns amino acid profile.

        List of dictionaries mapping amino acid to degeneracy (redundance)
        Indicies of list reflects position within protein

        Usage
        -----
        >> supE = GeneticCode(1, {'TAG': 'Q'})
        >> oligo = Oligonucleotide('NNKGCC', supE)
        >> oligo.get_amino_acid_profile()
            [{'A':2, 'C':1, D':1, 'E':1, 'F':1, ...},
             {'A':1                                }]
        """
        self.amino_acid_profile = [c.translate(dtype='decimal')
                                   for c in self.codons]
        return self.amino_acid_profile

    def get_amino_acid_degeneracy_profile(self):
        """
        Generates and returns amino acid degeneracy profile

        Usage
        -----
        >> supE = GeneticCode(1, {'TAG': 'Q'})
        >> oligo = Oligonucleotide('NNKAAA', supE)
        >> oligo.get_amino_acid_degeneracy_profile()
            [{1:11, 2:6 , 3:3},
             {1:1            }]

        Explanation of profile (for 'NNKAAA' oligo with supE translation)
        ----------------------
        The 0 index refers to the first amino acid (or codon) position in oligo
            This NNK codon encodes 20 (11 + 6 + 3) amino acids
                11 of these amino acids (CDEFHIKMNQ) are encoded by 1 codon
                 6                      (AGPQTV)                    2 codons
                 3                      (LRS)                       3 codons
        The 1 index refers to the next position, where 'AAA' encodes for 'K'
                 1 amino acid           (K)                         1 codon
        """

        self.amino_acid_degeneracy_profile = []
        for d in self.amino_acid_profile:
            self.amino_acid_degeneracy_profile.append(
                {v: list(d.values()).count(v) for v in set(d.values())})

        return self.amino_acid_degeneracy_profile

    def get_protein_degeneracy_table(self):
        """
        Generates and returns protein degeneracy profile

        Protein degeneracy is a dictionary where keys are all possible
        degeneracies of proteins derived from the degenerate oligonucleotide,
        and where values are the number of proteins in this group.

        Dictionary values reflect number of proteins within Oligonucleotid
        with a given degeneracy.

        Usage
        -----
        >> supE = GeneticCode(1, {'TAG': 'Q'})
        >> oligo = Oligonucleotide('NNK'*2, supE)
        >> oligo.get_protein_degeneracy_table()
            {1: 121,  # key=1*1, value=11*11
             2: 132,  # key=1*2, value=11*6*2 (*2 because key=1*2 and 2*1)
             3: 66,
             4: 36,
             6: 36,
             9: 9}

        Explanation of profile (for 'NNKNNK' oligo with supE translation)
        ----------------------
        121 proteins (CC, CD, CE, ...) have a degeneracy of 1
        131 proteins (AC, AD, AE, ...) have a degeneracy of 2
        ...
        There are 20*20 = 400 = (121 + 132 + 66 + 36 + 36 + 9) unique protein
        members of this oligonucleotide ensemble.
        Multiplying the degeneracy by the protein counts give the number of
        unique DNA counts
        """

        # for each df, create new dfs for each row of next df,
        # then concat, repeat
        dfs = [pd.DataFrame({'Degeneracy': list(_.keys()),
                             'Proteins': list(map(decimal.Decimal,
                                                  _.values()))},
                            dtype='object')
               for _ in self.amino_acid_degeneracy_profile]

        if not dfs:
            return pd.DataFrame()

        df = dfs[0]
        for next_df in dfs[1:]:
            df = pd.concat([df * next_df.iloc[i] for i in range(len(next_df))])
            df = df.groupby(['Degeneracy'], sort=False).sum().reset_index()

        self.protein_degeneracy_table = df.sort_values(by=['Degeneracy']) \
                                          .reset_index(drop=True)
        return self.protein_degeneracy_table

    def get_degeneracy_table(self):
        """
        Expands protein degeneracy table
        """
        df = self.protein_degeneracy_table.copy()
        df['Oligonucleotides'] = df.Degeneracy * df.Proteins

        df['DNA_Quantile'] = \
            df.Oligonucleotides.cumsum() / sum(df.Oligonucleotides)
        df['Protein_Quantile'] = \
            df.Proteins.cumsum() / sum(df.Proteins)
        self.degeneracy_table = df

        # defining additional aliases
        self.protein_quantiles = df['Protein_Quantile'].tolist()
        self.dna_quantiles = df['DNA_Quantile'].tolist()
        self.degeneracies = df['Degeneracy'].tolist()

        self.df = df

    def get_size(self, output='dna'):
        """
        Calculates and returns size of oligonucleotide ensemble
        """

        self.size_oligonucleotides = self.df['Oligonucleotides'].sum()
        self.size_proteins = self.df['Proteins'].sum()
        self.size = self.size_oligonucleotides

        if output.startswith('d'):
            return self.size_oligonucleotides
        elif output.startswith('p'):
            return self.size_proteins
        else:
            return (self.size_oligonucleotides, self.size_proteins)

    def get_gini_index(self):
        """
        Calculates Gini Index of protein ensemble.
        """
        x = [0] + self.protein_quantiles
        y = [0] + self.dna_quantiles
        n = len(self.protein_quantiles)

        gini = decimal.Decimal(1 / 2)
        for i in range(n):
            gini -= (x[i + 1] - x[i]) * (y[i + 1] + y[i]) / 2

        self.gini_index = 2 * gini
        return self.gini_index

    def get_makowski_diversity(self):
        """
        **INCORRECT**
        Calculates diversity as define by Makowski and Soares
        https://doi.org/10.1093/bioinformatics/btg013
        d = 1/(N*SUM(Pi^2))
        """
        d = decimal.Decimal(0)
        P = self.size_proteins

        degeneracy = self.degeneracy_table['Degeneracy']
        protein_counts = self.degeneracy_table['Proteins']

        for x, y in zip(protein_counts, degeneracy):
            p = y / self.size_oligonucleotides
            d += x * (p ** 2)

        self.makowski_diversity = 1 / (P * d)
        return self.makowski_diversity

    def _get_alternative_makowski_diversity(self):
        """
        EQUIVALENT ANSWER AS self.get_makowski_diversity()
        Calculates diversity as define by Makowski and Soares
        https://doi.org/10.1093/bioinformatics/btg013
        Equation (2)
        d = 1/(N*PRODUCTi{SUMj{Pij^2}})
        Where i is position protein,
              j is each amino acids,
              Pij is prob of amino acid j at i
        """
        d = decimal.Decimal(1)

        for aa_dict in self.amino_acid_profile:
            tot = sum(aa_dict.values())
            d *= sum([(aa / tot)**2 for aa in aa_dict.values()])

        self._makowski_diversity = 1 / (self.size_proteins * d)
        return self._makowski_diversity

    def get_entropy(self, base=None):
        """
        Calculates entropy of abundance profile
        s = -SUM(pi*ln(pi))
        """

        degeneracy = self.degeneracy_table['Degeneracy']
        protein_counts = self.degeneracy_table['Proteins']

        print(degeneracy)
        print(protein_counts)
        return 0
        entropy = decimal.Decimal(0)

        # for each degeneracy group
        for x, y in zip(protein_counts, degeneracy):
            p = y / self.size_proteins
            entropy += x * p * decimal.Decimal(math.log(p))

            if base is not None:
                entropy /= math.log(base)

        self.entropy = -entropy
        return self.entropy

    def __repr__(self):
        return f'Oligonucleotide({self.label})'


def reverse_complement(oligo):
    """
    """
    if isinstance(oligo, Oligonucleotide):
        oligo = oligo.label

    # TODO!


def get_nnk(n=1):
    """
    Simple helper function to easily generate NNK-based oligos.
    """
    supE = GeneticCode(1, {'TAG': 'Q'})
    oligo = Oligonucleotide('NNK' * n, supE)
    return oligo



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from dogma import (
    GeneticCode,
    DEFAULT_GENETIC_CODE,
    translate,
    Nucleotide,
    Codon,
)


class Oligonucleotide:
    """
    A sequence of nucleotides.
    """

    def __init__(self, data, genetic_code=None):
        """
        Multiple input formats are acceptable
            Oligonucleotide('AAANNK')
            Oligonucleotide(list('AAANNK'))
            Oligonucleotide(list(map(Nucleotide, list('AAANNK'))))
            Oligonucleotide([{'A':1}, {'A':0.5, 'C':0.5}]
            Oligonucleotide(list(map(Nucleotide, list('AAANNK'))))
        """

        if not isinstance(genetic_code, GeneticCode):
            genetic_code = DEFAULT_GENETIC_CODE
        self.genetic_code = genetic_code

        if isinstance(data, (str, list)):
            if isinstance(data[0], Codon):
                self.bases = []
                for c in data:
                    self.bases += c.bases
            else:
                self.bases = [_
                              if isinstance(_, Nucleotide)
                              else Nucleotide(_)
                              for _ in data]
        elif isinstance(data, Codon):
            self.bases = data.bases
            self.genetic_code = data.genetic_code
        else:
            self.bases = None

        self.length = self.get_length()
        self.label = self.get_label()
        self.compact_label = self.get_compact_label()

        self.codon_strings = self.get_codon_strings()
        self.codons = self.get_codons()

        self.get_base_profile()
        self.get_codon_profile()

    def get_bases(self):
        """
        Returns list of Nucleotides objects
        """
        return self.bases

    def get_codons(self):
        """
        Returns list of Codon objects derived from oligonucleotide bases
        """
        base_triplets = [self.bases[3 * i: 3 * i + 3]
                         for i in range(self.length // 3)]
        return [Codon(_, self.genetic_code) for _ in base_triplets]

    def get_length(self):
        """
        Returns integer length of oligonucleotide, number of bases
        """
        return len(self.bases)

    def get_label(self):
        """
        Returns a string of nucleotide labels.
        """
        return ''.join(_.label for _ in self.bases)

    def get_codon_strings(self):
        """
        Returns list of codon strings.
        """
        return [self.label[3 * i: 3 * i + 3] for i in range(self.length // 3)]

    def get_compact_label(self):
        """
        Returns shorthand label
        Attempts to reduce identical codon labels with numbers
        Ex: NNKNNKNNKAAANNK --> NNK<3>AAANNK
        """
        count = 1
        previous_string = ''
        output = []
        for c in self.get_codon_strings():
            if c == previous_string:
                count += 1
            else:
                if count > 1:
                    output.append(f'<{count}>')
                previous_string = c
                output.append(c)
                count = 1
        if count > 1:
            output.append(f'<{count}>')
        return ''.join(output)

    def translate(self, genetic_code=None):
        """
        Translates oligonucleotide into protein.

        Parameters:
        genetic_code: GeneticCode object
        """
        if not isinstance(genetic_code, GeneticCode):
            genetic_code = self.genetic_code
        return ''.join([genetic_code[_] for _ in self.get_codon_strings()])

    def samples(self, k=1, output='oligonucleotide', by_codon=True):
        return [self.sample(output=output, by_codon=by_codon)
                for _ in range(k)]

    def sample(self, output='oligonucleotide', by_codon=True):
        """
        Returns a single non-degenerate oligonucleotide string.

        If by_codon is True, sampling is based on the codons attribute,
        otherwise sampling is performed on each base individually.

        """
        if by_codon:
            oligonucleotide_string = ''.join(c.sample() for c in self.codons)
        else:
            oligonucleotide_string = ''.join([b.sample() for b in self.bases])

        if output == 'protein':
            return translate(oligonucleotide_string, self.genetic_code)
        elif output == 'both':
            return oligonucleotide_string, translate(oligonucleotide_string,
                                                     self.genetic_code)
        else:
            return oligonucleotide_string

    def get_base_profile(self):
        """
        Generates and returns base profile

        Usage
        -----
        >> nnk = Codon('NNK')
        >> oligo = Oligonucleotide(nnk)
        >> oligo.get_base_profile()
            {0: {'A':1, 'C':1 , 'G':1. 'T':1},
             1: {'A':1, 'C':1 , 'G':1. 'T':1},
             2: {               'G':1. 'T':1}}
        """
        self.base_profile = [b.composition for b in self.bases]
        return self.base_profile

    def __repr__(self):
        return f'Oligonucleotide({self.label})'


def reverse_complement(oligo):
    """
    """
    if isinstance(oligo, Oligonucleotide):
        oligo = oligo.label

    # TODO!


def get_nnk(n=1):
    """
    Simple helper function to easily generate NNK-based oligos.
    """
    supE = GeneticCode(1, {'TAG': 'Q'})
    oligo = Oligonucleotide('NNK' * n, supE)
    return oligo



from setuptools import setup, find_packages
from codecs import open
from os import path

__version__ = '0.0.2'

here = path.abspath(path.dirname(__file__))


DESCRIPTION = 'Collection of object-oriented entities related by the central dogma of biology.',
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

with open(path.join(here, 'requirements.txt'), encoding='utf-8') as f:
    all_reqs = f.read().split('\n')

install_requires = [x.strip() for x in all_reqs if 'git+' not in x]
dependency_links = [x.strip().replace('git+', '')
                    for x in all_reqs if x.startswith('git+')]

setup(
    name='dogma',
    version=__version__,
    description=DESCRIPTION,
    long_description=long_description,
    license='MIT',
    author='Griffin Clausen',
    author_email='griffinclausen@gmail.com',
    keywords='',
    entry_points='''
        [console_scripts]
        genetic_code=dogma.genetic_code:test
    ''',
    packages=find_packages(exclude=['docs', 'tests*']),
    include_package_data=True,
    install_requires=[
        'pandas'
        ],
    dependency_links=dependency_links
)



#!/usr/bin/env python
# -*-encoding=utf-8-_*-


"""Package-wide variables and helper functions."""

from itertools import product
from random import choice


DEFAULT_DECIMAL_PRECISION = 200

STANDARD_NUCLEOTIDES = list('ACGT')

DEGENERATE_NUCLEOTIDE_CODE = {
    'A': 'A',
    'C': 'C',
    'G': 'G',
    'T': 'T',
    'R': 'AG',
    'Y': 'CT',
    'S': 'CG',
    'W': 'AT',
    'K': 'GT',
    'M': 'AC',
    'B': 'CGT',
    'D': 'AGT',
    'H': 'ACT',
    'V': 'ACG',
    'N': 'ACGT'
}

DEGENERATE_NUCLEOTIDES = sorted(DEGENERATE_NUCLEOTIDE_CODE.keys())

DEGENERATE_NUCLEOTIDE_CODE_REVERSED = {v: k for k, v in
                                       DEGENERATE_NUCLEOTIDE_CODE.items()}

DEGENERATE_NUCLEOTIDE_CODE_COMPOSITION = {code: {n: int(n in _)
                                          for n in STANDARD_NUCLEOTIDES}
                                          for code, _ in
                                          DEGENERATE_NUCLEOTIDE_CODE.items()}

NUCLEOTIDE_BASE_PAIRS = dict(zip('ACGT', 'TGCA'))

_D = DEGENERATE_NUCLEOTIDE_CODE_REVERSED
DEGENERATE_NUCLEOTIDE_PAIRS = {k: _D[''.join(sorted(
                                     [NUCLEOTIDE_BASE_PAIRS[_]
                                      for _ in v]))]
                               for k, v in DEGENERATE_NUCLEOTIDE_CODE.items()}

DEFAULT_NUCLEOTIDE_LABEL = 'N'

DEGENERATE_NUCLEOTIDE_PAIRS_ = {
    'A': 'T',
    'C': 'G',
    'G': 'C',
    'T': 'A',
    'R': 'Y',
    'Y': 'R',
    'S': 'S',
    'W': 'W',
    'K': 'M',
    'M': 'K',
    'B': 'V',
    'D': 'H',
    'H': 'D',
    'V': 'B',
    'N': 'N'
}


DEFAULT_OLIGONUCLEOTIDE_LABEL = 'NNN'

STANDARD_CODONS = [''.join(_) for _ in
                   product(STANDARD_NUCLEOTIDES, repeat=3)]

DEGENERATE_CODONS = [''.join(_) for _ in
                     product(DEGENERATE_NUCLEOTIDES, repeat=3)]

DEFAULT_CODON_LABEL = 'NNN'

STANDARD_AMINO_ACIDS = 'ACDEFGHIKLMNPQRSTVWY'

DEFAULT_AMINO_ACID_LABEL = 'X'

STOP_LABEL = '*'

DEFAULT_RESIDUE_LABEL = 'X'


def rescale(data, total=1):
    """
    Rescales numerical values in lists or dictionary values to sum
    to specified total.

    Usage
    *****
    rescale([1, 3]) -> [0.25 0.75]
    rescale({'a': 1, 'b':'9']) -> {'a': 0.1, 'b': 0.9}
    """

    if isinstance(data, list):
        input_total = sum(data)
        assert input_total != 0, 'Error in doe.rescale(), input_total == 0'

        return [_ / input_total * total for _ in data]

    elif isinstance(data, dict):
        input_total = sum(data.values())
        assert input_total != 0, 'Error in doe.rescale(), input_total == 0'

        return {k: v / input_total * total for k, v in data.items()}
    else:
        return None


def get_frequency_dictionary(data):
    """
    Takes a string or list of strings and returns a dictionary of unique
    members and their abundance. Similar to itertools.Counter.
    """
    return {k: data.count(k) for k in set(data)}


def get_random_oligonucleotide(length=3, letters='ACGTRYSWKMBDHVN'):
    """
    Returns string of random degenerate nucleotides.

    Parameters
    **********
    length - int - length of string to return
    letters - string or list - collection of letters to select from, with replacement.

    Each of the 15 letters 'ACGTRYSWKMBDHVN' are equally likely.
    Alternatively, use letters='ACGT' for standard oligo strings
    """
    return ''.join(choice(list(letters)) for _ in range(length))
